import warnings
from typing import Any, AnyStr, Callable, Dict, List, Optional, Type, Union

import pandas as pd
from loguru import logger

from orkg.graph import subgraph
from orkg.logging_messages import MessageBuilder


def _filter_dict(
    d: Dict[str, Any],
    keys_to_exclude: Union[str, List[str]],
    required_key: Optional[str] = None,
) -> Union[Dict[str, Any], None]:
    """
    Returns a new dictionary that contains all the key-value pairs from `d`, except for those whose keys are in `keys_to_exclude`.

    If `required_key` is provided and it exists in `d`, the function will return a new dictionary that contains all the key-value pairs from `d`, except for those whose keys are in `keys_to_exclude`. If `required_key` is not provided or it does not exist in `d`, the function will return None.

    Args:
    - d: A dictionary whose keys are strings and whose values can be of any type.
    - keys_to_exclude: A string or list of strings representing the keys to be excluded from the output dictionary.
    - required_key: An optional string representing a key that must exist in `d` for the function to return a dictionary. If `required_key` is not provided or it does not exist in `d`, the function will return None.

    Returns:
    - A new dictionary that contains all the key-value pairs from `d`, except for those whose keys are in `keys_to_exclude`, or None if `required_key` is not provided or it does not exist in `d`.
    """
    if isinstance(keys_to_exclude, str):
        keys_to_exclude = [keys_to_exclude]
    if required_key is not None and required_key not in d:
        return None
    return {k: v for k, v in d.items() if k not in keys_to_exclude}


class OrkgResponse(object):
    # Class-level variable to store the ORKG client
    orkg_client = None

    # Class-level dictionary to store type conversion functions
    type_registry = {}

    @property
    def succeeded(self) -> bool:
        return str(self.status_code)[0] == "2"

    def __init__(
        self,
        client,
        response,
        status_code: Optional[str],
        content: Optional[Union[List, Dict, AnyStr]],
        url: str,
        paged: bool,
    ):
        self.orkg_client = client
        if response is None and status_code is None and content is None and url is None:
            logger.debug(MessageBuilder.backend_response_none())
            raise ValueError(
                "either response should be provided or content with status code"
            )
        if not paged:
            logger.warning(MessageBuilder.deprecation_warning("paged=False"))
            warnings.warn(
                "You are running an out-of-date ORKG backend! GET (i.e., listing) calls now provide pagination information",
                DeprecationWarning,
            )
        if (response is not None and str(response.status_code)[0] != "2") or (
            status_code is not None and content is not None and status_code[0] != "2"
        ):
            self.status_code = (
                response.status_code if response is not None else status_code
            )
            self.content = response.content if response is not None else content
            self.page_info = None
            logger.debug(
                MessageBuilder.backend_response(
                    self.content, self.status_code, response
                )
            )
            return
        if response is not None:
            self.status_code = response.status_code
            self.content = (
                response.json()
                if len(response.content) > 0
                and response.content.decode("utf-8")[0] in ["[", "{"]
                else response.content.decode("utf-8")
            )
            self.page_info = None
            if paged and isinstance(self.content, dict):
                self.page_info = _filter_dict(self.content, "content", "pageable")
                self.content = (
                    self.content["content"]
                    if "content" in self.content
                    else self.content
                )
            self.url = response.url
            logger.debug(
                MessageBuilder.backend_response(
                    self.content, self.status_code, self.url
                )
            )
        if status_code is not None and content is not None:
            self.status_code = status_code
            self.content = content
            self.page_info = None
            if paged and isinstance(self.content, dict) and "content" in content:
                self.page_info = _filter_dict(self.content, "content", "pageable")
                self.content = self.content["content"]
            self.url = url
            logger.debug(
                MessageBuilder.backend_response(
                    self.content, self.status_code, self.url
                )
            )

    def __repr__(self) -> str:
        return "%s %s" % ("(Success)" if self.succeeded else "(Fail)", self.content)

    def __str__(self) -> Union[List, Dict]:
        return self.content

    def as_type(self, object_type: Type, *args, **kwargs) -> Any:
        """
        Convert the content of the response to the specified type using the registered conversion function.
        Args:
            object_type (Type): a type (e.g., float, pandas.DataFrame).
            *args: Variable length argument list to pass to the conversion function.
            **kwargs: Keyword arguments to pass to the conversion function.

        Returns:
            T: The content of the response converted to the specified type.
        """
        if object_type not in self.type_registry:
            raise ValueError(f"Type '{object_type}' is not registered for conversion.")

        conversion_function = self.type_registry[object_type]
        return conversion_function(self, *args, **kwargs)

    @classmethod
    def register_type(cls, object_type: Type, conversion_function: Callable) -> None:
        """
        Register a type conversion function in the type registry.
        Args:
            object_type (Type): a type (e.g., float, pandas.DataFrame).
            conversion_function (Callable): Function that performs the conversion.
        """
        cls.type_registry[object_type] = conversion_function

    def as_float(self) -> float:
        """
        Convert the content of the response to a float.
        returns: a float
        """
        if self.content is None:
            raise ValueError("Response content is None")
        if isinstance(self.content, list):
            raise ValueError("Can't convert list to int")
        return self.as_type(float)

    def as_int(self) -> int:
        """
        Convert the content of the response to an int.
        returns: an int
        """
        if self.content is None:
            raise ValueError("Response content is None")
        if isinstance(self.content, list):
            raise ValueError("Can't convert list to int")
        return self.as_type(int)

    def as_dataframe(self) -> pd.DataFrame:
        """
        Convert the content of the response to a pandas DataFrame.
        returns: a pandas DataFrame
        """
        if self.content is None:
            raise ValueError("Response content is None")
        if isinstance(self.content, list):
            raise ValueError("Can't convert list to a dataframe")
        if self.content["_class"] != "resource":
            raise ValueError("Response content is not a resource")
        if "Table" not in self.content["classes"]:
            raise ValueError("resource is not a table object")
        nx_object = subgraph(client=self.orkg_client, thing_id=self.content["id"])
        columns = [
            nx_object.edges(s, data=True)
            for s in [
                o
                for s, o, p in nx_object.edges(self.content["id"], data=True)
                if p["id"] == "CSVW_Columns"
            ]
        ]
        columns_of_df = []
        for column in columns:
            name = ""
            order = -1
            for _, o, p in column:
                if p["id"] == "CSVW_Number":
                    order = nx_object.nodes[o]["label"]
                if p["id"] == "CSVW_Name":
                    name = nx_object.nodes[o]["label"]
            columns_of_df.append((int(order), name))
        rows = [
            nx_object.edges(s, data=True)
            for s in [
                o
                for s, o, p in nx_object.edges(self.content["id"], data=True)
                if p["id"] == "CSVW_Rows"
            ]
        ]
        result = [[] for _ in range(len(rows))]
        index_of_df = []
        for row in rows:
            order = -1
            title = ""
            values = []
            for s, o, p in row:
                if p["id"] == "CSVW_Number":
                    order = nx_object.nodes[o]["label"]
                if p["id"] == "CSVW_Titles":
                    title = nx_object.nodes[o]["label"]
                if p["id"] == "CSVW_Cells":
                    cells = nx_object.edges(o, data=True)
                    for _, o, p in cells:
                        if p["id"] == "CSVW_Value":
                            values.append(nx_object.nodes[o]["label"])
                        else:
                            values.append("")
            index_of_df.append((int(order), title))
            result[int(order) - 1] = values
        columns = [c[1] for c in sorted(columns_of_df, key=lambda x: x[0])]
        index = [i[1] for i in sorted(index_of_df, key=lambda x: x[0])]
        return pd.DataFrame(result, columns=columns, index=index)


# Register conversion functions, order doesn't matter
OrkgResponse.register_type(float, lambda response: float(response.content["label"]))
OrkgResponse.register_type(int, lambda response: int(response.content["label"]))
OrkgResponse.register_type(pd.DataFrame, OrkgResponse.as_dataframe)


class OrkgUnpaginatedResponse(object):
    @property
    def all_succeeded(self) -> bool:
        return all([response.succeeded for response in self.responses])

    def __init__(self, responses: List[OrkgResponse]):
        self.responses = responses
        self.content: List[Dict] = []

        for response in self.responses:
            self.content.extend(response.content)



from __future__ import unicode_literals

import warnings
from datetime import date, datetime
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Tuple, Union
from urllib.parse import quote_plus, urlencode

from loguru import logger

from orkg.logging_messages import MessageBuilder
from orkg.out import OrkgResponse, OrkgUnpaginatedResponse

if TYPE_CHECKING:
    from orkg.client.backend import ORKG  # Forward declaration only for type checking

# parts of URL to be omitted
SKIP_IN_PATH = (None, "", b"", [], ())


def _escape(value: Any) -> Any:
    """
    Escape a single value of a URL string or a query parameter. If it is a list
    or tuple, turn it into a comma-separated string first.
    """

    # dates and datetimes into isoformat
    if isinstance(value, (date, datetime)):
        value = value.isoformat()

    # make bools into true/false strings
    elif isinstance(value, bool):
        value = str(value).lower()

    # don't decode bytestrings
    elif isinstance(value, bytes):
        return value

    if isinstance(value, (list, tuple)):
        return value
    return str(value)


def _make_path(*parts: Union[List, Tuple]) -> str:
    """
    Create a URL string from parts, omit all `None` values and empty strings.
    Convert lists and tuples to comma separated values.
    """
    return "/" + "/".join(
        # preserve ',' and '*' in url for nicer URLs in logs
        quote_plus(_escape(p), b",*")
        for p in parts
        if p not in SKIP_IN_PATH
    )


# parameters that apply to all methods
GLOBAL_PARAMS = ("pretty", "human", "error_trace", "format", "filter_path")


def query_params(*query_params):
    """
    Decorator that pops all accepted parameters from method's kwargs and puts
    them in the params argument.
    """

    def _wrapper(func):
        @wraps(func)
        def _wrapped(*args, **kwargs):
            params = {}
            if "params" in kwargs:
                params = kwargs.pop("params").copy()
            for p in query_params + GLOBAL_PARAMS:
                if p in kwargs:
                    v = kwargs.pop(p)
                    if v is not None:
                        params[p] = _escape(v)

            # don't treat ignore and request_timeout as other params to avoid escaping
            for p in ("ignore", "request_timeout"):
                if p in kwargs:
                    params[p] = kwargs.pop(p)
            return func(*args, params=params, **kwargs)

        return _wrapped

    return _wrapper


def dict_to_url_params(params) -> str:
    return "?%s" % urlencode(params)


def simcomp_available(func):
    def check_if_simcomp_available(self: NamespacedClient, *args, **kwargs):
        if not self.client.simcomp_available:
            raise ValueError("simcomp_host must be provided in the ORKG wrapper class!")
        return func(self, *args, **kwargs)

    return check_if_simcomp_available


def admin_functionality(func):
    def display_admin_warning(self: NamespacedClient, *args, **kwargs):
        # TODO: create some functionality in the backend to check for permissions
        warnings.warn("This call needs elevated role in the system!", RuntimeWarning)
        return func(self, *args, **kwargs)

    return display_admin_warning


def check_host(func):
    """
    a decorator to check whether the backend is live or not.
    """

    def ping_decorator(self, *args, **kwargs):
        if not self.client.ping():
            raise ValueError("The ORKG is down!")
        return func(self, *args, **kwargs)

    return ping_decorator


class NamespacedClient(object):
    client: "ORKG"
    auth: Optional[dict]

    def __init__(self, client: "ORKG"):
        self.client = client
        if self.client.token is not None:
            self.auth = {"Authorization": f"Bearer {self.client.token}"}
        else:
            self.auth = None

    def __getattribute__(self, name: str) -> Any:
        """
        Logs any method called in the class or its children and the arguments passed to it
        """
        attr = object.__getattribute__(self, name)
        if not name.startswith("__") and callable(attr):

            def log_method_calls(*args, **kwargs):
                error = False
                # log when method is first called
                logger.debug(MessageBuilder.method_call(attr, name, args, kwargs))
                try:
                    return attr(*args, **kwargs)
                except Exception as e:
                    error = True
                    raise e
                finally:
                    if not error:
                        # log when method execution is complete
                        logger.debug(MessageBuilder.execution_complete(attr, name))

            return log_method_calls
        else:
            # return the attribute normally
            return attr

    @staticmethod
    def handle_sort_params(params: dict) -> None:
        if "sort" in params:
            sort_direction = "asc"
            if "desc" in params:
                sort_direction = "desc" if params["desc"] else "asc"
                del params["desc"]
            params["sort"] = f"{params['sort']},{sort_direction}"

    @staticmethod
    def _call_pageable(
        func: Callable[..., OrkgResponse],
        args: dict,
        params: dict,
        start_page: int = 0,
        end_page: int = -1,
    ) -> OrkgUnpaginatedResponse:
        def _call_one_page(page_number: int) -> Tuple[int, OrkgResponse]:
            try:
                params["page"] = page_number
                response = func(**args, params=params)
            except TypeError:
                # FIXME: check whether "func" is pageable without calling it !?
                raise ValueError(
                    'Provided "func" is either not pageable or the passed arguments are wrong'
                )

            if not response.succeeded:
                return 0, response

            if not response.page_info or "totalPages" not in response.page_info:
                raise ValueError('Provided "func" is not pageable')

            return response.page_info["totalPages"], response

        pages = []
        n_pages, page = _call_one_page(page_number=start_page)
        pages.append(page)

        if -1 < end_page < n_pages:
            n_pages = end_page

        for i in range(start_page + 1, n_pages):
            _, page = _call_one_page(page_number=i)
            pages.append(page)

        return OrkgUnpaginatedResponse(responses=pages)


def verb_logger(verb: str) -> Callable:
    def decorator(func: Callable) -> Callable:
        def wrapper(*args, **kwargs):
            if kwargs:
                logger.debug(
                    MessageBuilder.api_call_with_request_body(verb, args, kwargs)
                )
            else:
                logger.debug(MessageBuilder.api_call_simple(verb, args))

            return func(*args, **kwargs)

        return wrapper

    return decorator



"""
Some imports here though are not used by the codegen.py file itself, but are used by the generated code.
So they should be kept. and in case of any linter support in the future, they should be ignored.
"""

import datetime
import types
from typing import Dict, List, Optional, Tuple, Union

import pandas as pd
from cardinality import at_least, at_most
from inflector import English, Inflector

from orkg.client.templates.components import DataFrame, NestedTemplate, Template
from orkg.client.templates.out import ObjectStatement, TemplateInstance
from orkg.client.templates.utils import (
    Boolean,
    Date,
    Integer,
    Number,
    String,
    check_against_keywords,
    clean_up_dict,
    post_process_dict,
    pre_process_string,
    stringify_value,
)
from orkg.common import OID
from orkg.utils import NamespacedClient


class OTFFunctionWriter(object):
    inflector = Inflector(English)

    @staticmethod
    def convert_df_to_list_of_lists(df: pd.DataFrame, label: Optional[str] = None) -> Dict[str, Union[str, List[List]]]:
        result = [[c for c in df.columns]]
        for index, row in df.iterrows():
            result.append([row[c] for c in df.columns])
        to_return = {'@df': result}
        if label is not None:
            to_return['label'] = label
        return to_return

    @staticmethod
    def implement_function(
            orkg_context: NamespacedClient,
            template_id: str
    ):
        template = Template.find_or_create_template(orkg_context.client, template_id)
        params, params_docstring = template.get_params_as_function_params()
        object_json, properties = template.create_api_object()
        lookup_map = {component.get_clean_name(): component.predicate_id for component in template.components}
        values_fillers_variable = "partial_json"
        values_fillers = OTFFunctionWriter.create_values_fillers(properties, lookup_map, values_fillers_variable)
        cardinality_checks = OTFFunctionWriter.create_cardinality_checks(template)
        function_name = check_against_keywords(pre_process_string(template.template_name))
        new_method = f'''
def {function_name}(self, {params}) -> TemplateInstance:
    """ Creates a template of type {template_id} ({template.template_name})

    {params_docstring}
    :return: a string representing the resource ID of the newly created resource
    """
    lookup_map = {lookup_map}
{cardinality_checks}
    obj = ObjectStatement.create_main({'""' if template.is_formatted else 'label'}, ['{template.template_class}'])
    {values_fillers_variable} = {object_json}
{values_fillers}
    obj = {values_fillers_variable}
    for param_name, nested_template in {{k: v for k, v in locals().items() if any(isinstance(v, t) for t in (TemplateInstance, pd.DataFrame, Tuple, OID))}}.items():
        predicate_id = lookup_map[param_name]
        if 'values' not in obj['resource']:
            obj['resource']['values'] = {{}}
        if predicate_id not in obj['resource']['values']:
            obj['resource']['values'][predicate_id] = []
        if isinstance(nested_template, pd.DataFrame):
            obj['resource']['values'][predicate_id].append(OTFFunctionWriter.convert_df_to_list_of_lists(nested_template))
        elif isinstance(nested_template, Tuple):
            obj['resource']['values'][predicate_id].append(OTFFunctionWriter.convert_df_to_list_of_lists(*nested_template))
        elif isinstance(nested_template, OID):
            obj['resource']['values'][predicate_id].append(nested_template.to_orkg_json())
        else:
            obj['resource']['values'][predicate_id].append(clean_up_dict(nested_template.template_dict))
    post_process_dict(obj)
    return TemplateInstance(obj, self.client)

if '{function_name}' in orkg_context.client.templates.materialized_templates:
    delattr(orkg_context.client.templates, '{function_name}')
orkg_context.client.templates.{function_name} = types.MethodType( {function_name}, orkg_context )
orkg_context.client.templates.materialized_templates.add('{function_name}')
                    '''
        exec(new_method)

    @staticmethod
    def create_cardinality_checks(template: Template) -> str:
        code = ""
        for component in template.components:
            param_name = component.get_clean_name()
            if component.min_cardinality or component.max_cardinality:
                error_message = f"The parameter {param_name} must have"
                if component.min_cardinality:
                    error_message += f" at least {component.min_cardinality} {OTFFunctionWriter.inflector.conditional_plural(component.min_cardinality, 'value')}" if component.min_cardinality else ""
                if len(error_message) > 0 and component.max_cardinality:
                    error_message += f" and at most {component.max_cardinality} {OTFFunctionWriter.inflector.conditional_plural(component.max_cardinality, 'value')}" if component.max_cardinality else ""
                code += f"    if isinstance({param_name}, List):\n"
                if component.min_cardinality:
                    code += f"        if not at_least({component.min_cardinality}, {param_name})"
                if component.max_cardinality:
                    if component.min_cardinality:
                        code += f" or not at_most({component.max_cardinality}, {param_name})"
                    else:
                        code += f"        if not at_most({component.max_cardinality}, {param_name})"
                code += ":\n"
                code += f"            raise ValueError('Oops! {error_message}')\n"
        return code

    @staticmethod
    def create_values_fillers(properties: List[ObjectStatement], properties_lookup_map: Dict, values_fillers_variable: str) -> str:
        code = ""
        properties_dict = {p.predicate_id: p for p in properties}
        for param_name, predicate_id in properties_lookup_map.items():
            if predicate_id not in properties_dict:
                continue
            object_description = properties_dict[predicate_id].serialize()
            if len(object_description) > 0:
                code += f"\n    # Updating {param_name}\n"
                code += f"    if '{predicate_id}' not in {values_fillers_variable}:\n"
                code += f"        {values_fillers_variable}['resource']['values']['{predicate_id}'] = []\n"
                code += f"    if isinstance({param_name}, OID):\n"
                code += f"        {values_fillers_variable}['resource']['values']['{predicate_id}'].append({{'@id': str({param_name})}})\n"
                code += f"    elif isinstance({param_name}, List):\n"
                code += f"        for i in range(len({param_name})):\n"
                code += f"            if isinstance({param_name}[i], OID):\n"
                code += f"                object_description_at_i = {{'@id': str({param_name}[i])}}\n"
                code += f"            object_description_at_i = {object_description}\n"
                code += f"            if 'text' in object_description_at_i:\n"
                code += f"                object_description_at_i['text'] = {param_name}[i]\n"
                code += f"            elif 'label' in object_description_at_i:\n"
                code += f"                object_description_at_i['label'] = {param_name}[i]\n"
                code += f"            {values_fillers_variable}['resource']['values']['{predicate_id}'].append(object_description_at_i)\n"
                code += f"    else:\n"
                code += f"        {values_fillers_variable}['resource']['values']['{predicate_id}'].append({object_description})\n"
        return code



import sys
from typing import Union

from loguru import logger

LEVELS = (
    "TRACE",
    "DEBUG",
    "INFO",
    "SUCCESS",
    "WARNING",
    "ERROR",
    "CRITICAL",
    5,
    10,
    20,
    25,
    30,
    40,
    50,
)


def initialize_logger(logging_level: Union[str, int]) -> None:
    """
    Creates a logger for each ORKG instance.
    Defines the format, content, and colors of the logging messages and the output location.
    :param logging_level: the minimum severity threshold from which log messages are shown
    """
    if logging_level not in LEVELS:
        raise ValueError(
            "the provided logging level is not valid! Options: 'DEBUG', 'INFO', 'WARNING', 'ERROR'"
        )
    logger.remove()
    logger.level("INFO", color="<light-blue>")
    logger.level("DEBUG", color="<fg #4c73c2>")
    fmt = "<fg #65bfea>[{time:HH:mm:ss.SS}]</fg #65bfea> <level>[{level}]</level> <fg #4cc295>[{name}:{function}]</fg #4cc295> <fg #b3f5f3>{message}</fg #b3f5f3>"
    logger.add(sys.stderr, level=logging_level, colorize=True, format=fmt)



from typing import Dict, Optional

from orkg.out import OrkgResponse, OrkgUnpaginatedResponse
from orkg.utils import NamespacedClient, dict_to_url_params, query_params


class ClassesClient(NamespacedClient):
    def by_id(self, id: str) -> OrkgResponse:
        self.client.backend._append_slash = True
        response = self.client.backend.classes(id).GET()
        return self.client.wrap_response(response)

    @query_params("q", "exact")
    def get_all(self, params: Optional[Dict] = None) -> OrkgResponse:
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend.classes.GET(dict_to_url_params(params))
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.classes.GET()
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc", "q", "exact")
    def get_resource_by_class(
        self, class_id: str, params: Optional[Dict] = None
    ) -> OrkgResponse:
        self.handle_sort_params(params)
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend.classes(class_id).resources.GET(
                dict_to_url_params(params)
            )
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.classes(class_id).resources.GET()
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc", "q", "exact")
    def get_resource_by_class_unpaginated(
        self,
        class_id: str,
        start_page: int = 0,
        end_page: int = -1,
        params: Optional[Dict] = None,
    ) -> OrkgUnpaginatedResponse:
        return self._call_pageable(
            self.get_resource_by_class,
            args={"class_id": class_id},
            params=params,
            start_page=start_page,
            end_page=end_page,
        )

    @query_params("id", "label", "uri")
    def add(self, params: Optional[Dict] = None) -> OrkgResponse:
        if len(params) == 0:
            raise ValueError("at least label should be provided")
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.classes.POST(json=params, headers=self.auth)
        return self.client.wrap_response(response)

    @query_params("id", "label", "uri")
    def find_or_add(self, params: Optional[Dict] = None) -> OrkgResponse:
        if len(params) == 0:
            raise ValueError("at least label should be provided")
        else:
            if "id" in params:
                # check if a class with this id is there
                found = self.by_id(params["id"])
                if found.succeeded:
                    return found
            # check if a class with this label is there
            found = self.get_all(q=params["label"], exact=True)
            if found.succeeded:
                if isinstance(found.content, list) and len(found.content) > 0:
                    found.content = found.content[0]
                    return found
            # None found! let's create a new one
            self.client.backend._append_slash = True
            response = self.client.backend.classes.POST(json=params, headers=self.auth)
        return self.client.wrap_response(response)

    @query_params("label", "uri")
    def update(self, id: str, params: Optional[Dict] = None) -> OrkgResponse:
        if len(params) == 0:
            raise ValueError("at least label should be provided")
        else:
            if not self.exists(id):
                raise ValueError("the provided id is not in the graph")
            self.client.backend._append_slash = True
            response = self.client.backend.classes(id).PUT(
                json=params, headers=self.auth
            )
        return self.client.wrap_response(response)

    def exists(self, id: str) -> bool:
        return self.by_id(id).succeeded



from typing import Dict, List, Union

from orkg.out import OrkgResponse
from orkg.utils import NamespacedClient


class DummyClient(NamespacedClient):
    def create_xxx_response(
        self, code: str, content: Union[Dict, List]
    ) -> OrkgResponse:
        return OrkgResponse(
            client=self,
            status_code=code,
            content=content,
            paged=False,
            response=None,
            url="",
        )

    def create_200_response(self, content: Union[Dict, List]) -> OrkgResponse:
        return self.create_xxx_response("200", content)

    def create_404_response(self, content: Union[Dict, List]) -> OrkgResponse:
        return self.create_xxx_response("404", content)



from typing import Dict, List, Optional, Tuple

from orkg.common import ComparisonType, ThingType
from orkg.out import OrkgResponse
from orkg.utils import NamespacedClient


class ComparisonsClient(NamespacedClient):
    def publish(
        self,
        contribution_ids: Optional[List[str]] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
        reference: Optional[str] = None,
        comparison_method: ComparisonType = ComparisonType.PATH,
        predicates: Optional[List[str]] = None,
        transposed: bool = False,
        data: Optional[Dict] = None,
    ) -> OrkgResponse:
        """
        Create a comparison resource in backend and a corresponding thing reference in SimComp
        :param contribution_ids: a list of contributions' ids to compare (optional)
        :param title: the title of the comparison
        :param description: the description of the comparison (optional)
        :param reference: the reference of the comparison (optional)
        :param comparison_method: the method used to compare the contributions - PATH or MERGE (default: PATH)
        :param predicates: the list predicates to strict the comparison view (optional)
        :param transposed: whether to transpose the comparison view (default: False)
        :param data: the data of the comparison - should be valid data according to SimComp specifications (optional)
        :return: the response of the request
        """
        if title is None:
            raise ValueError("title should be provided")

        if contribution_ids is None and data is None:
            raise ValueError(
                "either contribution_ids or precomputed data should be provided"
            )

        description, reference = self._validate_comparison_components(
            contribution_ids, description, reference
        )
        comparison_id, comparison_response = self._create_comparison_in_graph(
            title, description, reference, contribution_ids
        )

        self._save_comparison_to_simcomp(
            comparison_id,
            contribution_ids,
            comparison_method,
            predicates,
            transposed,
            data,
        )

        return comparison_response

    def _create_comparison_in_graph(
        self,
        title: str,
        description: str,
        reference: str,
        contribution_ids: Optional[List[str]],
    ) -> Tuple[str, OrkgResponse]:
        """
        Create a comparison resource in ORKG
        :param title: the title of the comparison
        :param description: the description of the comparison
        :param reference: the reference of the comparison
        :param contribution_ids: a list of contributions' ids to compare (optional)
        :return: the id of the comparison and the response of the request
        """
        # create the comparison resource
        comparison_response = self.client.resources.add(
            label=title, classes=["Comparison"]
        )
        # create the literals
        comparison_id = comparison_response.content["id"]
        description_id = self.client.literals.add(label=description).content["id"]
        reference_id = self.client.literals.add(label=reference).content["id"]
        # add the statements
        self.client.statements.add(
            subject_id=comparison_id,
            predicate_id="description",
            object_id=description_id,
        )
        self.client.statements.add(
            subject_id=comparison_id, predicate_id="reference", object_id=reference_id
        )
        # If a list of contributions is provided, add them to the comparison
        if contribution_ids is not None:
            for contribution_id in contribution_ids:
                self.client.statements.add(
                    subject_id=comparison_id,
                    predicate_id="compareContribution",
                    object_id=contribution_id,
                )
        return comparison_id, comparison_response

    def _validate_comparison_components(
        self,
        contribution_ids: Optional[List[str]],
        description: Optional[str],
        reference: Optional[str],
    ) -> Tuple[str, str]:
        """
        Validate the components of the comparison
        :param contribution_ids: a list of contributions' ids to compare
        :param description: the description of the comparison (optional)
        :param reference: the reference of the comparison (optional)
        :return: the validated description and reference of the comparison
        """
        if contribution_ids is not None:
            for contribution_id in contribution_ids:
                contribution_resource = self.client.resources.by_id(
                    id=contribution_id
                ).content

                if "Contribution" not in contribution_resource["classes"]:
                    raise ValueError(
                        "this ID is not a contribution: " + str(contribution_id)
                    )
        description = description if description is not None else ""
        reference = reference if reference is not None else ""
        return description, reference

    def _save_comparison_to_simcomp(
        self,
        comparison_id: str,
        contribution_ids: List[str],
        comparison_type: ComparisonType,
        predicates: Optional[List[str]],
        transposed: bool,
        data: Optional[Dict],
    ):
        """
        Save the comparison to SimComp. Either computes a fresh comparison or uses the provided data
        :param comparison_id: the id of the comparison
        :param contribution_ids: a list of contributions' ids to compare
        :param comparison_type: the method used to compare the contributions - PATH or MERGE
        :param predicates: the list predicates that customize the comparison view (optional)
        :param transposed: whether to transpose the comparison view
        :param data: the data of the comparison - should be valid data according to SimComp specifications (optional)
        """
        if data is None:
            # Compute a fresh comparison
            data = self.client.contributions.compare(
                contributions=contribution_ids,
                comparison_type=comparison_type,
            ).content["payload"]["comparison"]
        # Use the provided data
        config = {
            "contributions": contribution_ids,
            "transpose": transposed,
            "type": comparison_type.value.upper().strip(),
        }
        if predicates is not None:
            config["predicates"] = predicates
        self.client.json.save_json(
            thing_key=comparison_id,
            thing_type=ThingType.COMPARISON,
            data=data,
            config=config,
        )



from typing import Callable, Dict, List, Optional, Tuple, Union


class MessageBuilder:
    @staticmethod
    def method_call(attr: Callable, method_name: str, args: Tuple, kwargs: Dict) -> str:
        """
        Returns message for logging a call to a method (with module path) and arguments passed to it

        :param attr: method object
        :param method_name: name of method
        :param args: the positional arguments passed to the method
        :param kwargs: the keyword arguments passed to the method
        :return: logging message
        """
        arg_str = MessageBuilder._format_args(args, kwargs)
        func = (
            f"{attr.__module__}.{method_name}"
            if hasattr(attr, "__module__")
            else f"{method_name}"
        )
        return f"Calling: {func}({arg_str})"

    @staticmethod
    def execution_complete(attr: Callable, method_name: str) -> str:
        """
        Returns message for logging that a method has finished executing

        :param attr: method object
        :param method_name: name of method
        """
        func = (
            f"{attr.__module__}.{method_name}"
            if hasattr(attr, "__module__")
            else f"{method_name}"
        )
        return f"Execution complete: {func}"

    @staticmethod
    def api_call_simple(verb: str, args: Tuple) -> str:
        """
        Returns message for logging API calls

        :param verb: the HTTP method
        :param args: url and parameters for the api request
        :return: logging message
        """
        return f"API-{verb} {'/'.join([str(param) for param in args])}"

    @staticmethod
    def api_call_with_request_body(verb: str, args: Tuple, kwargs: Dict) -> str:
        """
        Returns message for logging API calls

        :param verb: the HTTP method
        :param args: url and parameters for the api request
        :param kwargs: body of the request
        :return: logging message
        """
        return f"API-{verb} {'/'.join([str(param) for param in args])} body: { {} if 'json' not in kwargs else kwargs['json']}"

    @staticmethod
    def backend_response(
        content: Union[Dict, List], status_code: Union[int, str], url: Optional[str]
    ) -> str:
        """
        Returns message for logging the response from the backend
        """
        size = len(content) if type(content) is list else 1
        content = content[0:5] if type(content) is list else content
        return f"Backend response - status code: {status_code}, content length: {size}, content (max 5 objects): {content}, url: {url}"

    @staticmethod
    def backend_response_none() -> str:
        return "No response from backend!"

    @staticmethod
    def deprecation_warning(feature: str) -> str:
        """
        Returns message for logging usage of a deprecated feature

        :param feature: parameter/method/variable/etc. which is deprecated
        :return: logging message
        """
        return f"Deprecated feature usage: {feature}"

    @staticmethod
    def _format_args(args: Tuple[Union[str, Dict, Callable]], kwargs: Dict) -> str:
        """
        Concatenates positional and keyword arguments into a single string

        :param args: the positional arguments passed to the method
        :param kwargs: the keyword arguments passed to the method
        :return: string of all arguments
        """
        arg_str = ""
        if args:
            if isinstance(args[0], str):
                arg_str += ", ".join([str(a) for a in args])
            elif isinstance(args[0], dict):
                arg_str += MessageBuilder._stringify_dict(arg_str, args[0])
            elif callable(args[0]):
                arg_str += f"{args[0].__module__}.{args[0].__name__}"
            else:
                print("args type:", type(args[0]))
        if kwargs:
            arg_str = MessageBuilder._stringify_dict(arg_str, kwargs)
        return arg_str

    @staticmethod
    def _stringify_dict(string: str, dict_: Dict) -> str:
        """
        Converts dict into string formatted as 'key1 = val1, key2 = val2'
        """
        for k, v in dict_.items():
            if string:
                string += f", {k} = {v}"
            else:
                string += f"{k} = {v}"
        return string



from typing import Dict, Optional

from orkg.out import OrkgResponse
from orkg.utils import NamespacedClient, dict_to_url_params, query_params


class LiteralsClient(NamespacedClient):
    def by_id(self, id: str) -> OrkgResponse:
        self.client.backend._append_slash = True
        response = self.client.backend.literals(id).GET()
        return self.client.wrap_response(response)

    @query_params("q", "exact")
    def get_all(self, params: Optional[Dict] = None) -> OrkgResponse:
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend.literals.GET(dict_to_url_params(params))
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.literals.GET()
        return self.client.wrap_response(response)

    @query_params("id", "label", "datatype")
    def add(self, params: Optional[Dict] = None) -> OrkgResponse:
        if len(params) == 0:
            raise ValueError("at least label must be provided")
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.literals.POST(json=params, headers=self.auth)
        return self.client.wrap_response(response)

    @query_params("label", "datatype")
    def update(self, id: str, params: Optional[Dict] = None) -> OrkgResponse:
        if len(params) == 0:
            raise ValueError("label must be provided")
        else:
            if not self.exists(id):
                raise ValueError("the provided id is not in the graph")
            self.client.backend._append_slash = True
            response = self.client.backend.literals(id).PUT(
                json=params, headers=self.auth
            )
        return self.client.wrap_response(response)

    def exists(self, id: str) -> bool:
        return self.by_id(id).succeeded



from collections import defaultdict
from typing import Dict, List, Optional, Tuple, Union

import pandas as pd

from orkg.common import ComparisonType, ExportFormat, ThingType
from orkg.out import OrkgResponse
from orkg.utils import NamespacedClient, simcomp_available


class ContributionsClient(NamespacedClient):
    @simcomp_available
    def similar(self, contribution_id: str, candidate_count: int = 10) -> OrkgResponse:
        """
        Get contributions which are similar to a specified contribution.
        Returns a list of [contribution ID, contribution label, paper ID, paper title, and similarity score].

        :param contribution_id: contribution ID
        :param candidate_count: number of similar contributions to return (default: 10)
        :return: OrkgResponse object of similar contributions
        """
        self.client.simcomp._append_slash = True
        response = self.client.simcomp.contribution.similar.GET(
            params={"contribution_id": contribution_id, "n_results": candidate_count}
        )
        return self.client.wrap_response(response=response)

    @simcomp_available
    def compare(
        self,
        contributions: List[str],
        comparison_type: ComparisonType = ComparisonType.PATH,
        export_format: Optional[ExportFormat] = None,
    ) -> OrkgResponse:
        """
        Get comparison by list of contribution IDs.

        :param contributions: list of contribution IDs
        :param comparison_type: type of comparison to perform (default: ComparisonType.PATH)
        :param export_format: format of comparison export (default: None)
        :return: OrkgResponse object of the comparison
        """
        self.client.simcomp._append_slash = False
        response = self.client.simcomp.contribution.compare.GET(
            params={
                "contributions": contributions,
                "type": comparison_type.value,
                "format": None if export_format is None else export_format.value,
            }
        )
        return self.client.wrap_response(response=response)

    @simcomp_available
    def compare_dataframe(
        self,
        contributions: Optional[List[str]] = None,
        comparison_id: Optional[str] = None,
        like_ui=True,
        include_meta=False,
        comparison_type: ComparisonType = ComparisonType.PATH,
        thing_type: Optional[ThingType] = ThingType.COMPARISON,
    ) -> Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]:
        """
        Fetches an ORKG comparison into a pandas.DataFrame.
        Optional: create a second DataFrame of contribution metadata.
        Included metadata (when present): author, doi, publication month, publication year, url, research field, venue

        :param contributions: list of contribution IDs from a comparison
        :param comparison_id: ID of a comparison
        :param like_ui: true/false whether to match comparison df to its UI representation
        :param include_meta: true/false whether to return metadata df
        :param comparison_type: the method used to compare the contributions - PATH or MERGE (default: PATH)
        :param thing_type: type of thing to export (default: COMPARISON)
        :return:
            - comparison DataFrame
            - metadata DataFrame (optional)
        """
        # Check if valid request
        if contributions is None and comparison_id is None:
            raise ValueError("either provide the contributions, or the comparison ID")
        if contributions is not None and comparison_id is not None:
            raise RuntimeWarning(
                "both contributions and comparison ID provided, using comparison ID"
            )

        if comparison_id is not None:
            df: pd.DataFrame = self._export_comparison(
                comparison_id, like_ui, thing_type
            )
        else:
            df: pd.DataFrame = self._compare_contributions_and_export(
                comparison_type, contributions
            )

        if include_meta:
            # Collect required components to generate metadata dataframe
            if comparison_id is not None:
                raw_response = self.client.json.get_json(
                    thing_key=comparison_id, thing_type=thing_type
                ).content
            else:
                raw_response = self.compare(
                    contributions=contributions, comparison_method=comparison_type
                ).content
            # Use helper method to create metadata df
            return df, self._create_metadata_df(
                [c for c in df.columns],
                {
                    contribution[
                        "id"
                    ]: f"{contribution['paper_label']}/{contribution['label']}"
                    for contribution in raw_response["payload"]["thing"]["data"][
                        "contributions"
                    ]
                },
                raw_response["payload"]["thing"]["data"]["contributions"],
            )
        return df

    def _compare_contributions_and_export(
        self, comparison_type: ComparisonType, contributions: List[str]
    ) -> pd.DataFrame:
        """
        Compare contributions and return a DataFrame.
        :param comparison_type: the method used to compare the contributions - PATH or MERGE (default: PATH)
        :param contributions: list of contribution IDs from a comparison
        :return: a pandas DataFrame
        """
        # Use the compare endpoint
        response = self.compare(
            contributions=contributions,
            comparison_type=comparison_type,
            export_format=ExportFormat.DATAFRAME,
        )
        if not response.succeeded:
            raise RuntimeError(
                f"Failed to fetch live data for contributions {contributions}"
            )
        df = self._convert_simcomp_df_to_pandas_df(response.content)
        return df

    def _export_comparison(
        self, comparison_id: str, like_ui: bool, thing_type: ThingType
    ) -> pd.DataFrame:
        """
        Export a comparison and return a DataFrame.
        :param comparison_id: ID of a comparison
        :param like_ui: true/false whether to match comparison df to its UI representation
        :param thing_type: type of thing to export (default: COMPARISON)
        :return: a pandas DataFrame
        """
        # Use the export endpoint
        response = self.client.simcomp.thing.export.GET(
            params={
                "thing_key": comparison_id,
                "thing_type": thing_type.value,
                "format": ExportFormat.DATAFRAME.value,
                "like_ui": like_ui,
            }
        )
        if not response.ok:
            raise RuntimeError(f"Failed to get thing {comparison_id}")
        df = self._convert_simcomp_df_to_pandas_df(response.json())
        return df

    @staticmethod
    def _convert_simcomp_df_to_pandas_df(simcomp_json: Dict) -> pd.DataFrame:
        df = pd.DataFrame.from_dict(simcomp_json)
        # iterate over each cell in the dataframe
        for column in df.columns:
            for row in range(len(df)):
                # if a value contains <SEP> split it and make the cell a list
                if "<SEP>" in df[column][row]:
                    df[column][row] = df[column][row].split("<SEP>")
        return df

    def _create_metadata_df(
        self,
        columns: List[str],
        contribution_ids_and_titles: Dict[str, str],
        contributions_list: List[Dict],
    ) -> pd.DataFrame:
        """
        Create a metadata DataFrame containing the following properties (when available):
            author, doi, publication month, publication year, url, research field, venue
        Column headings are identical to the comparison DataFrame.

        :param columns: list of headings for df columns
        :param contribution_ids_and_titles: dictionary of contribution IDs to headings: Paper Title/Contribution N (Contribution ID)
        :param contributions_list: list of contribution dictionaries containing contribution label, ID, paper ID, title, and year
        :return: metadata DataFrame
        """
        paper_set = set(
            [f"{contribution['paper_id']}" for contribution in contributions_list]
        )
        comparison_meta_data = {}
        # create dict of meta properties and values
        for paper in paper_set:
            contribution_ids, paper_meta_data = self._get_paper_metadata(
                contributions_list, paper
            )
            # add paper meta info for each contribution to comparison meta info
            for contribution_id in contribution_ids:
                column_name = contribution_ids_and_titles[contribution_id]
                paper_dict = paper_meta_data.copy()
                comparison_meta_data[column_name] = paper_dict
                comparison_meta_data[column_name]["contribution id"] = contribution_id
        # Make dataframe with same column order as comparison df and replace missing fields with empty string
        df_meta = pd.DataFrame.from_dict(comparison_meta_data)[columns].fillna("")
        return df_meta

    def _get_paper_metadata(
        self, contributions_list: List[Dict], paper: str
    ) -> Tuple[List, Dict]:
        """
        Get metadata info for a paper.

        :param contributions_list: list of contribution dictionaries containing contribution label, ID, paper ID, title, and year
        :param paper: ID of the paper
        :return:
            - list of IDs of contributions belonging to this paper
            - dictionary of (available) paper metadata properties to values
        """
        # author, doi, publication month, publication year, url, research field, venue
        meta_property_ids = ["P27", "P26", "P28", "P29", "url", "P30", "HAS_VENUE"]
        paper_meta_dict_of_lists = defaultdict(list)
        paper_statements = self.client.statements.get_by_subject_unpaginated(
            subject_id=paper
        )
        for statement in paper_statements.content:
            if statement["predicate"]["id"] in meta_property_ids:
                pred = statement["predicate"]["label"]
                obj = statement["object"]["label"]
                paper_meta_dict_of_lists[pred].append(obj)
        # make dict values strings if only one and list of strings if two or more
        paper_meta_data = {
            k: v.pop() if len(v) == 1 else v
            for k, v in paper_meta_dict_of_lists.items()
        }
        paper_meta_data["title"] = paper_statements.content[0]["subject"]["label"]
        paper_meta_data["paper id"] = paper
        # get IDs of all contributions from this paper which are included in the comparison
        contribution_ids = [
            str(contribution["id"])
            for contribution in contributions_list
            if contribution["paper_id"] == paper
        ]
        return contribution_ids, paper_meta_data



import json
import os
import re
from typing import Any, Dict, List, Optional, Union

import requests

from orkg.client.harvesters.utils import process_contribution
from orkg.common import OID
from orkg.out import OrkgResponse


def _validate_doi(doi: str) -> bool:
    """Check if a string is a valid DOI or a complete DOI URL."""
    if doi is None:
        return False

    # DOI pattern.
    doi_pattern = re.compile(r"^10.\d{4,9}/[-._;()/:A-Z0-9]+$", re.I)

    # DOI URL pattern.
    url_pattern = re.compile(r"^https?://doi\.org/10.\d{4,9}/[-._;()/:A-Z0-9]+$", re.I)

    return bool(doi_pattern.match(doi) or url_pattern.match(doi))


def harvest(
    orkg_client: Any,
    doi: Optional[str],
    orkg_rf: Union[str, OID],
    directory: Optional[str],
) -> OrkgResponse:
    if doi is None and directory is None:
        raise ValueError("Either doi or directory must be provided.")

    contributions_urls = []
    doi_response = None

    # Check if directory is provided and all is valid
    if directory is not None:
        if not os.path.isdir(directory):
            raise ValueError(f"The directory {directory} does not exist.")
        if doi is None and not os.path.isfile(os.path.join(directory, "doi.json")):
            raise ValueError(
                f"The directory {directory} does not contain the file doi.json."
            )
        if os.path.isfile(os.path.join(directory, "doi.json")):
            with open(os.path.join(directory, "doi.json")) as f:
                doi_response = json.load(f)
    if doi_response is None:
        doi_response, contributions_urls = _get_doi_response(doi)

    if isinstance(orkg_rf, str):
        rf_response = orkg_client.resources.get(q=orkg_rf, exact=True, size=1)
        if not rf_response.succeeded or len(rf_response.content) == 0:
            raise ValueError(
                f"Unable to find the ORKG research field with the given string value {orkg_rf}"
            )
        orkg_rf = OID(rf_response.content[0]["id"])

    doi_content = doi_response["data"]
    paper_json = {
        # TODO: handle multiple titles
        "title": doi_content["attributes"]["titles"][0]["title"],
        "doi": doi_content["attributes"]["doi"],
        "authors": [
            {"label": f'{creator["givenName"]} {creator["familyName"]}'}
            for creator in doi_content["attributes"]["creators"]
        ],
        "publicationYear": int(doi_content["attributes"].get("publicationYear"))
        if doi_content["attributes"].get("publicationYear")
        else None,
        "publishedIn": doi_content["attributes"].get("publisher", None),
        "researchField": f"{orkg_rf}",
    }

    if "contributions" not in paper_json:
        paper_json["contributions"] = []

    # Get the contribution content and override if directory is provided
    contribution_content = [requests.get(url).json() for url in contributions_urls]
    contribution_content = _filter_to_orkg_harvestable_content(contribution_content)

    if directory is not None:
        files = [
            f
            for f in os.listdir(directory)
            if os.path.isfile(os.path.join(directory, f)) and f != "doi.json"
        ]
        contribution_content: List[Dict] = []
        for file in files:
            with open(os.path.join(directory, file)) as f:
                contribution_content.append(json.load(f))

    for contribution_json in contribution_content:
        orkg_contribution_json = {}
        global_ids = {}
        context = contribution_json.get("@context", {})
        process_contribution(
            contribution_json, orkg_contribution_json, global_ids, context
        )
        # replace the key "label" with "name"
        orkg_contribution_json["name"] = orkg_contribution_json.pop("label")
        paper_json["contributions"].append(orkg_contribution_json)

    # Remove @temp from all the contribution objects (NamedObject doesn't have a @temp field anymore)
    for contribution in paper_json["contributions"]:
        if "@temp" in contribution:
            del contribution["@temp"]

    # Now that we have everything, let's finalize the paper object and add it to the graph
    paper_json = {"paper": paper_json}
    return orkg_client.papers.add(paper_json)


def _filter_to_orkg_harvestable_content(contribution_content: List[Dict]) -> List[Dict]:
    new_contribution_content = []
    for contribution in contribution_content:
        if (
            "@id" in contribution
            and "label" in contribution
            and "@type" in contribution
            and any("orkg.org/class" in t for t in contribution["@type"])
        ):
            new_contribution_content.append(contribution)
    return new_contribution_content


def _get_doi_response(doi: str):
    # Check if the doi is a valid DOI string
    if not _validate_doi(doi):
        raise ValueError(f"{doi} is not a valid DOI string")

    # Strip doi from https://doi.org/ if present
    doi = (
        doi.replace("https://doi.org/", "")
        if doi.startswith("https://doi.org/")
        else doi
    )

    # Get the supplementary content via datacite
    # pass the doi part only
    url = f"https://api.datacite.org/dois?query=relatedIdentifiers.relatedIdentifier:{doi}"
    response = requests.get(url, headers={"Accept": "application/json"})
    if response.status_code != 200:
        raise ValueError(
            f"Unable to retrieve the supplementary content via reverse lookup of {doi}"
        )
    response = response.json()

    if "data" in response and len(response["data"]) == 0:
        raise ValueError(f"No supplementary material found via reverse lookup of {doi}")

    response["data"] = response["data"][
        0
    ]  # TODO: Handle multiple supplementary materials hits

    # Check that this file is meant as a supplementary material to our DOI
    if not any(
        [
            ri["relatedIdentifier"] == doi
            for ri in response["data"]["attributes"]["relatedIdentifiers"]
            if ri["relationType"] == "IsSupplementTo"
        ]
    ):
        raise ValueError(
            f"Supplementary material {response['data']['id']} is not a related to {doi}"
        )

    # get contribution info
    contributions_urls = [
        url["relatedIdentifier"]
        for url in filter(
            lambda x: x["relationType"] == "HasPart"
            and x["relatedIdentifierType"] == "URL"
            and x["resourceTypeGeneral"] == "Dataset",
            response["data"]["attributes"]["relatedIdentifiers"],
        )
    ]
    return response, contributions_urls



from typing import Dict

from orkg.out import OrkgResponse
from orkg.utils import NamespacedClient


class ObjectsClient(NamespacedClient):
    def add(self, params: Dict) -> OrkgResponse:
        """
        Warning: Super-users only should use this endpoint
        Create a new object in the ORKG instance
        :param params: orkg Object
        :return: an OrkgResponse object containing the newly created object resource
        """
        self.client.backend._append_slash = True
        response = self.client.backend.objects.POST(json=params, headers=self.auth)
        return self.client.wrap_response(response)



import os
import re
from ast import literal_eval
from typing import Dict, List, Optional, Tuple, Union

from deprecated import deprecated
from pandas import read_csv

from orkg.out import OrkgResponse
from orkg.utils import NamespacedClient, query_params


class PapersClient(NamespacedClient):
    def add(
        self, params: Optional[Dict] = None, merge_if_exists: bool = False
    ) -> OrkgResponse:
        """
        Create a new paper in the ORKG instance
        :param params: paper Object
        :param merge_if_exists: merge the papers if they exist in the graph and append contributions
        :return: an OrkgResponse object containing the newly created paper resource
        """
        self.client.backend._append_slash = True
        response = self.client.backend.papers.POST(
            json=params, params={"mergeIfExists": merge_if_exists}, headers=self.auth
        )
        return self.client.wrap_response(response)

    @query_params("file", "standard_statements")
    @deprecated(
        version="0.16.3",
        reason="The CSV import functionality will be removed in a future update :( Please use our lovely frontend instead!",
    )
    def add_csv(self, params: Optional[Dict] = None) -> OrkgResponse:
        """
        Create a new paper in the ORKG instance
        :param file: csv file containing the papers, with at least a column with title: "paper:title"
        :param params: string representation of dict (e.g. "{'name': 1}") with default statements, should contain CSV_PLACEHOLDER (optional)
        :return: OrkgResponse object of the added paper
        """
        if "file" not in params:
            raise ValueError(
                "missing file parameter!, please pass the path to a csv file"
            )
        if not os.path.exists(params["file"]):
            raise ValueError(f"file path doesn't exist!, path={params['file']}")
        df = read_csv(params["file"], dtype=str)
        contribution_ids = []
        for index in range(len(df)):
            paper = df.iloc[index].to_dict()
            # at least a paper title should be present
            if "paper:title" not in paper:  # TODO: enable crossref lookup based on DOI
                print('Skipping paper, column "paper:title" not found')
                continue
            # extract and format paper info
            (
                paper,
                paper_metadata,
                standard_statements,
            ) = PapersClient._get_paper_statements_and_metadata(paper, params)
            # generate the statements
            research_problems, contribution_statements = self._extract_statements(paper)
            # add default statements if they are present
            # TODO: only works for one level now, make it recursive
            statements_to_insert = standard_statements.copy()
            if len(statements_to_insert) > 0:
                contribution_statements = self._insert_standard_statements(
                    contribution_statements, statements_to_insert
                )
            # add research problem to the contribution
            if len(research_problems) > 0:
                contribution_statements = self._insert_research_problem(
                    contribution_statements, research_problems
                )
            # check if a paper with this title already exists
            existing_papers = self.client.resources.get(
                q=paper_metadata["title"], exact=True
            ).content
            existing_paper_id = 0
            for existing_paper in existing_papers:
                if "Paper" in existing_paper["classes"]:
                    existing_paper_id = existing_paper["id"]
                    break
            # paper does not yet exist
            if existing_paper_id == 0:
                paper_metadata["contributions"] = [
                    {"name": "Contribution 1", "values": contribution_statements}
                ]
                paper = {
                    "paper": paper_metadata  # now the full paper, not just metadata
                }
                contribution_ids = self._add_new_paper(contribution_ids, paper)
            # paper exists already, so add a contribution
            else:
                contribution_ids = self._add_new_contribution(
                    contribution_ids, contribution_statements, existing_paper_id
                )
        return self.client.wrap_response(content=contribution_ids, status_code="201")

    def by_doi(self, doi: str) -> OrkgResponse:
        """
        Get the paper resource in the ORKG associated with the given DOI.
        :param doi: The DOI of the paper, could be a full url or just the DOI
        :return: A list of paper objects
        """
        # Disable appending a slash to the base URL
        self.client.backend._append_slash = False

        # Extract the DOI from the input if it's a full URL
        if "doi.org" in doi:
            doi = doi.split("doi.org/")[-1]

        # Send an HTTP GET request to fetch paper information
        response = self.client.backend.papers.GET(params={"doi": doi})

        return self.client.wrap_response(response)

    def by_title(self, title: str) -> OrkgResponse:
        """
        Retrieve a paper object based on its title.
        :param title: The title of the research paper.
        :return: A list of paper objects
        """
        # Disable appending a slash to the base URL
        self.client.backend._append_slash = False

        # Send an HTTP GET request to fetch paper information
        response = self.client.backend.papers.GET(params={"title": title})

        return self.client.wrap_response(response)

    @staticmethod
    def _get_paper_statements_and_metadata(
        paper: Dict, params: Dict
    ) -> Tuple[Dict, Dict, Dict]:
        """
        Format paper statement values and add default values for missing data.
        Separate into dictionaries for statements and metadata.

        :param paper: dict of all info extracted from csv file
        :param params: string representation of dict (e.g. "{'name': 1}") dict with default statements, should contain CSV_PLACEHOLDER (optional)
        :return:
            - dict of paper statements (without metadata)
            - dict of paper metadata
            - dict of standard statements or empty dict if no standard statements
        """
        title = paper["paper:title"]
        authors = (
            [
                {"label": author.strip()}
                for author in str(paper["paper:authors"]).split(";")
            ]
            if "paper:authors" in paper
            and paper["paper:authors"] == paper["paper:authors"]
            else []
        )
        publication_month = paper.get("paper:publication_month", 1)
        publication_year = paper.get("paper:publication_year", 2000)
        research_field = paper.get("paper:research_field", "R11")  # "Science"
        doi = paper.get("paper:doi", "")
        url = ""
        published_in = ""
        paper_metadata = {
            "title": title,
            "authors": authors,
            "publicationMonth": publication_month,
            "publicationYear": publication_year,
            "researchField": research_field,
            "doi": doi,
            "url": url,
            "publishedIn": published_in,
        }
        standard_statements = (
            literal_eval(params["standard_statements"])
            if "standard_statements" in params
            else {}
        )
        # remove metadata from paper dict, already added above
        metadata_headers = [
            "paper:title",
            "paper:authors",
            "paper:publication_month",
            "paper:publication_year",
            "paper:doi",
            "paper:research_field",
        ]
        [paper.pop(key) if key in paper else paper for key in metadata_headers]
        return paper, paper_metadata, standard_statements

    def _add_new_paper(self, contribution_ids: List, paper: Dict) -> List:
        """
        Add new paper to ORKG instance and add contribution ID to list.

        :param contribution_ids: list of contribution ID strings
        :param paper: dict of paper statements and metadata
        :return: updated list of contribution ID strings
        """
        response = self.add(paper)
        if "id" in response.content:
            paper_id = response.content["id"]
            paper_statements = self.client.statements.get_by_subject(
                subject_id=paper_id, size=10000
            ).content
            for statement in paper_statements:
                if statement["predicate"]["id"] == "P31":
                    contribution_ids.append(statement["object"]["id"])
                    print("Added paper:", str(paper_id))
        else:
            print("Error adding paper: ", str(response.content))
        return contribution_ids

    def _add_new_contribution(
        self,
        contribution_ids: List,
        contribution_statements: Dict[str, List],
        existing_paper_id: str,
    ) -> List:
        """
        Add contribution to existing paper in ORKG instance and add contribution ID to list.

        :param contribution_ids: list of contribution ID strings
        :param contribution_statements: dict of predicate IDs and their objects
        :param existing_paper_id: paper ID string
        :return: updated list of contribution ID strings
        """
        paper_statements = self.client.statements.get_by_subject(
            subject_id=existing_paper_id, size=10000
        ).content
        contribution_amount = 0
        for paper_statement in paper_statements:
            if paper_statement["predicate"]["id"] == "P31":  # "Contribution"
                contribution_amount += 1
        contribution_id = self.client.resources.add(
            label="Contribution " + str(contribution_amount + 1),
            classes=["Contribution"],
        ).content["id"]
        self.client.statements.add(
            subject_id=existing_paper_id,
            predicate_id="P31",
            object_id=contribution_id,
        )
        self._create_statements(contribution_id, contribution_statements)
        contribution_ids.append(contribution_id)
        print(
            "Added contribution:",
            str(contribution_id),
            "to paper:",
            str(existing_paper_id),
        )
        return contribution_ids

    def _insert_research_problem(
        self, contribution_statements: Dict[str, List], research_problems: List[str]
    ) -> Dict[str, List]:
        """
        Add research problem(s) by resource ID to contribution statements dictionary.

        :param contribution_statements: dict of predicate IDs and their objects
        :param research_problems: list of (one or many) research problem(s) as strings
        :return: updated dict of contribution statements
        """
        contribution_statements["P32"] = []
        for research_problem in research_problems:
            research_problem_id = self.client.resources.find_or_add(
                label=research_problem, classes=["Problem"]
            ).content["id"]
            # P32 has research problem
            contribution_statements["P32"].append({"@id": research_problem_id})
        return contribution_statements

    def _insert_standard_statements(
        self, contribution_statements: Dict[str, List], statements_to_insert: Dict
    ) -> Dict[str, Union[str, List]]:
        """
        Create new standard statements dict with predicate IDs as keys and predicate labels as values.

        :param contribution_statements: dict of predicate IDs and their objects
        :param statements_to_insert: dict with default statements, should contain CSV_PLACEHOLDER
        :return: dict of standard statements' predicate IDs to predicate labels
        """
        for predicate in statements_to_insert:
            if isinstance(statements_to_insert[predicate], list):  # if is array
                for i in range(len(statements_to_insert[predicate])):
                    if (
                        statements_to_insert[predicate][i]["values"]
                        == "CSV_PLACEHOLDER"
                    ):
                        statements_to_insert[predicate][i][
                            "values"
                        ] = contribution_statements
            if not re.search("^P+[a-zA-Z0-9]*$", predicate):
                predicate_id = self.client.predicates.find_or_add(
                    label=predicate
                ).content["id"]
                statements_to_insert[predicate_id] = statements_to_insert[predicate]
                del statements_to_insert[predicate]
        return statements_to_insert

    def _extract_statements(self, paper: Dict) -> Tuple[List[str], Dict]:
        """
        Create a dictionary of predicate-object pairs formatted/typed for the ORKG.
        Predicates: find an existing ID matching the label, or create a new one.
        Objects: if resource, find an existing ID matching the label, or create a new one. Otherwise, add as literal.
        Also creates a list of research problem label(s).

        :param paper: dict of paper statements as extracted from CSV
        :return:
            - list of research problem label(s)
            - dict of predicate IDs and their objects
        """
        contribution_statements = {}
        research_problems = []
        for predicate in paper:
            value = paper[predicate]
            # add research problem (one or more)
            if predicate.startswith("contribution:research_problem"):
                research_problem = paper.get("predicate", "")
                if research_problem != "":
                    research_problems.append(research_problem)
                continue
            # filter out nan values
            if value != value:
                continue
            # to make columns unique, pandas appends a dot and number to duplicate columns, remove this here
            predicate = re.sub(r"\.[1-9]+$", "", predicate)
            value_is_resource = False
            # if predicate starts with resource:, insert it as resource instead of literal
            if predicate.startswith("resource:"):
                value_is_resource = True
                predicate = predicate[len("resource:") :]
            predicate_id = self.client.predicates.find_or_add(label=predicate).content[
                "id"
            ]
            if not value_is_resource:
                if predicate_id in contribution_statements:
                    contribution_statements[predicate_id].append({"text": value})
                else:
                    contribution_statements[predicate_id] = [{"text": value}]
            else:
                resource_id = self.client.resources.find_or_add(label=value).content[
                    "id"
                ]
                if predicate_id in contribution_statements:
                    contribution_statements[predicate_id].append({"@id": resource_id})
                else:
                    contribution_statements[predicate_id] = [{"@id": resource_id}]
        return research_problems, contribution_statements

    def _create_statements(self, subject_id: str, statements: Dict):
        """
        Add statements to ORKG instance.

        :param subject_id: string of subject resource ID
        :param statements: dict of predicate IDs and their objects
        """
        for predicate_id in statements:
            values = statements[predicate_id]
            for value in values:
                if "text" in value:
                    literal_id = self.client.literals.add(label=value["text"]).content[
                        "id"
                    ]
                    self.client.statements.add(
                        subject_id=subject_id,
                        predicate_id=predicate_id,
                        object_id=literal_id,
                    )
                elif "@id" in value:
                    self.client.statements.add(
                        subject_id=subject_id,
                        predicate_id=predicate_id,
                        object_id=value["@id"],
                    )
                elif "label" in value:
                    resource_id = self.client.resources.add(
                        label=value["label"]
                    ).content["id"]
                    self.client.statements.add(
                        subject_id=subject_id,
                        predicate_id=predicate_id,
                        object_id=resource_id,
                    )
                    self._create_statements(resource_id, value["values"])



from typing import Dict, Optional

from orkg.out import OrkgResponse
from orkg.utils import NamespacedClient, dict_to_url_params, query_params


class ResearchProblemsClient(NamespacedClient):
    @query_params("page", "size", "sort", "desc")
    def in_research_field(
        self, research_field_id: str, params: Optional[Dict] = None
    ) -> OrkgResponse:
        """
        Get all problems in a research field
        :param research_field_id: the id of the research field
        :param page: the page number (optional)
        :param size: number of items per page (optional)
        :param sort: key to sort on (optional)
        :param desc: true/false to sort desc (optional)
        :return: OrkgResponse object
        """
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend("research-fields")(
                research_field_id
            ).problems.GET(dict_to_url_params(params))
        else:
            self.client.backend._append_slash = True
            response = self.client.backend("research-fields")(
                research_field_id
            ).problems.GET()
        return self.client.wrap_response(response)



from typing import List, Optional


class ObjectStatement(object):
    """
    A class representing a statement in the ORKG object format.
    The object statement could represent a resource, a literal, or a nested resource.
    """

    # Optionals
    label: Optional[str]
    classes: Optional[List[str]]
    text: Optional[str]
    datatype: Optional[str]
    values: Optional[List]
    name: Optional[str]
    # Must have
    predicate_id: str
    template_id: str

    def __init__(self, predicate_id: str, template_id: str):
        self.predicate_id = predicate_id
        self.template_id = template_id
        self.literals = []
        self.resources = []
        self.values = None
        self.name = None

    @staticmethod
    def create_main(name: str, classes: List[str] = None) -> "ObjectStatement":
        statement = ObjectStatement("", "")
        statement.set_main_statement(name, classes)
        return statement

    @property
    def is_main_statement(self) -> bool:
        return self.name is not None

    def set_main_statement(self, name: str, classes: List[str] = None):
        self.name = name
        self.classes = classes

    def set_literal(self, text: str, datatype: Optional[str] = None):
        self.literals.append((text, datatype))

    def set_nested_statement(self, statement: "ObjectStatement"):
        self.set_resource(statement.label, statement.classes)
        for value in statement.values:
            self.add_value(value)

    def set_resource(self, label: str, classes: List[str] = None):
        self.resources.append((label, classes))

    def add_value(self, value: "ObjectStatement"):
        if self.values is None:
            self.values = []
        self.values.append(value)

    def values_to_statements(self) -> str:
        if self.values is None:
            return "{}"
        return (
            "{"
            + ", ".join(
                [f'"{v.predicate_id}": [{{{v.to_statement()}}}]' for v in self.values]
            )
            + "}"
        )

    def to_statement(self) -> str:
        statement = ""
        for literal in self.literals:
            text, datatype = literal
            statement += f'{{"text": stringify_value({text})'
            if datatype is not None:
                statement += f', "datatype": "{datatype}"'
            statement += "}"
        for resource in self.resources:
            label, classes = resource
            statement += f'{{"label": {label}'
            if classes is not None:
                statement += f', "classes": {classes}'
            statement += "}"
            if self.values is not None:
                statement += f', "values": {self.values_to_statements()} '
        return statement

    def serialize(self) -> str:
        if self.is_main_statement:
            return f'{{ "resource": {{ "name": "{self.name}", "classes": {self.classes}, "values": {self.values_to_statements()} }} }}'
        else:
            return f"{self.to_statement()}"



import datetime
import keyword
import re
from collections.abc import MutableMapping
from typing import Any, Dict, List, Union

Number = Union[int, float, complex, List[int], List[float], List[complex]]
Boolean = Union[bool, List[bool]]
String = Union[str, List[str]]
Integer = Union[int, List[int]]
Date = Union[datetime.date, List[datetime.date]]


def clean_up_dict(dictionary: dict) -> dict:
    """
    replaces the key 'name' with 'label' inside the 'resource' object.
    then strips the root key 'resource' and returns the inner dictionary.
    """
    dictionary = dictionary["resource"]
    dictionary["label"] = dictionary["name"]
    del dictionary["name"]
    return dictionary


def pre_process_string(string: str) -> str:
    lower_string = string.lower().replace(" ", "_")
    res = re.sub(r"\W+", "", lower_string)
    return res


def check_against_keywords(string: str) -> str:
    """
    Checks a string against python keywords.
    If a keyword is found, it is postfixed with a '_template'.
    """
    if string in keyword.kwlist:
        return string + "_template"
    return string


def map_type_to_pythonic_type(value_type: str) -> str:
    if value_type == "Number" or value_type == "Decimal":
        return "Number"
    elif value_type == "Boolean":
        return "Boolean"
    elif value_type == "Integer":
        return "Integer"
    elif value_type == "Date":
        return "Date"
    else:
        return "String"


def remove_empty_values(dictionary: dict):
    if "values" not in dictionary["resource"]:
        return
    for _, value in dictionary["resource"]["values"].items():
        for v in value:
            if not bool(v):
                value.remove(v)


def _check_duplicates(items: List[Dict[str, Any]]) -> None:
    """
    Remove duplicate dictionaries and dictionaries with None text or label
    from a list.

    Args:
        items (List[Dict[str, Any]]): List of dictionaries to check.
    """
    seen = []
    to_remove = []
    for item in items:
        if item in seen:
            to_remove.append(item)
        elif "text" in item and item["text"] == "None":
            to_remove.append(item)
        elif "label" in item and item["label"] == "None":
            to_remove.append(item)
        else:
            seen.append(item)
    # Remove duplicate or None text/label items
    for item in to_remove:
        items.remove(item)


def recursive_dict_check(input_dict: Dict[str, Any]) -> None:
    """
    Recursively traverse the dictionary to remove duplicate dictionaries
    and dictionaries with None text or label in all lists.

    Args:
        input_dict (Dict[str, Any]): Input dictionary.
    """
    for key, value in input_dict.items():
        if isinstance(value, list):
            # check for duplicate dicts in lists
            _check_duplicates(value)
            # also check each dict item in the list
            for item in value:
                if isinstance(item, MutableMapping):
                    recursive_dict_check(item)
        elif isinstance(value, MutableMapping):
            # if value is a dict, check the dict
            recursive_dict_check(value)


def remove_empty_lists_from_values(input_dict: Dict[str, Any]) -> None:
    """
    Recursively traverse the dictionary to remove empty lists from the 'values' key.

    Args:
        input_dict (Dict[str, Any]): Input dictionary.
    """
    to_remove = []
    for key, value in input_dict.items():
        if key == "values":
            for sub_key, sub_value in value.items():
                if isinstance(sub_value, list) and not sub_value:
                    to_remove.append(sub_key)
        elif isinstance(value, MutableMapping):
            remove_empty_lists_from_values(value)
    # remove empty lists
    for key in to_remove:
        del input_dict["values"][key]


def post_process_dict(dictionary: dict):
    remove_empty_values(dictionary)
    recursive_dict_check(dictionary)
    remove_empty_lists_from_values(dictionary)


def stringify_value(value: Any) -> str:
    if isinstance(value, datetime.date):
        return value.isoformat()
    else:
        return str(value)



from typing import Dict, List, Optional, Set, Tuple, Union

import pandas as pd

from orkg.client.templates.out import TemplateInstance
from orkg.client.templates.statements import ObjectStatement
from orkg.client.templates.utils import (
    check_against_keywords,
    map_type_to_pythonic_type,
    pre_process_string,
)
from orkg.common import OID
from orkg.utils import NamespacedClient

NATIVE_TYPES = ["Number", "String", "Boolean", "Integer", "Date", "URI", "Decimal"]
DATAFRAME_CLASSES = ["QBDataset", "Table"]

NestedTemplate = Union[TemplateInstance, OID, List[TemplateInstance], List[OID]]
DataFrame = Union[
    pd.DataFrame,
    Tuple[pd.DataFrame, str],
    OID,
    List[pd.DataFrame],
    List[Tuple[pd.DataFrame, str]],
    List[OID],
]


class TemplateComponent:
    """
    A template component, which represents a property of a template.
    The shape of each property is defined by the component.
    """

    predicate_id: str
    predicate_label: str
    value_class_id: Optional[str]
    value_class_label: Optional[str]
    is_of_custom_type: bool
    min_cardinality: Optional[int]
    max_cardinality: Optional[int]
    is_nested_template: bool
    nested_template_target: Optional[str]

    # private properties
    _orkg_client: NamespacedClient

    def __init__(self, orkg_client, component_id: str):
        self._orkg_client = orkg_client
        component_statements = orkg_client.statements.get_by_subject(
            subject_id=component_id, size=99999
        ).content
        # Set predicate info
        predicate = list(
            filter(lambda x: x["predicate"]["id"] == "sh:path", component_statements)
        )[0]["object"]
        self.predicate_id = predicate["id"]
        self.predicate_label = predicate["label"]
        # Set class value, id, and label
        value_class = list(
            filter(
                lambda x: x["predicate"]["id"] in ["sh:class", "sh:datatype"],
                component_statements,
            )
        )
        if len(value_class) == 0:
            self.value_class_id = "String"
            self.value_class_label = "Text"
            self.is_of_custom_type = False
        else:
            value_class = value_class[0]["object"]
            self.value_class_id = value_class["id"]
            self.value_class_label = value_class["label"]
            self.is_of_custom_type = (
                True if self.value_class_id not in NATIVE_TYPES else False
            )
        # Set cardinality of the property
        max_cardinality = list(
            filter(
                lambda x: x["predicate"]["id"] == "sh:maxCount", component_statements
            )
        )
        self.max_cardinality = (
            None
            if len(max_cardinality) == 0
            else int(max_cardinality[0]["object"]["label"])
        )
        min_cardinality = list(
            filter(
                lambda x: x["predicate"]["id"] == "sh:minCount", component_statements
            )
        )
        self.min_cardinality = (
            None
            if len(min_cardinality) == 0
            else int(min_cardinality[0]["object"]["label"])
        )
        # Set if the component is nested template
        if not self.is_of_custom_type:
            self.is_nested_template = False
        else:
            template_classes = orkg_client.statements.get_by_object_and_predicate(
                object_id=self.value_class_id, predicate_id="sh:targetClass", size=99999
            ).content
            if len(template_classes) == 0:
                self.is_nested_template = False
            else:
                self.is_nested_template = True
                self.nested_template_target = template_classes[0]["subject"]["id"]
                # Recursively materialize nested templates of the root
                if self.nested_template_target not in Template.wip_templates:
                    Template.wip_templates.add(self.nested_template_target)
                    self._orkg_client.templates.materialize_template(
                        self.nested_template_target
                    )

    def get_clean_name(self) -> str:
        return pre_process_string(check_against_keywords(self.predicate_label))

    def get_return_type(self, optional: bool = False) -> str:
        if not optional and self.min_cardinality == 0:
            return f"Optional[{self.get_return_type(True)}]"
        # Check if the return type should be a pandas.DataFrame
        if self.value_class_id in DATAFRAME_CLASSES:
            return "DataFrame"
        # Check if it is a nested template
        if self.is_nested_template:
            return "NestedTemplate"
        # o/w it is a pure python type
        return f"Union[{map_type_to_pythonic_type(self.value_class_id)}, OID]"

    def get_property_as_function_parameter(self) -> str:
        clean_label = self.get_clean_name()
        return_type = self.get_return_type()
        default_value = " = None" if self.min_cardinality == 0 else ""
        return f"{clean_label}: {return_type}{default_value}"

    def get_param_documentation(self) -> str:
        if self.is_nested_template:
            target_template = Template.find_or_create_template(
                self._orkg_client, self.nested_template_target
            )
            param_desc = f":param {self.get_clean_name()}: a nested template, use orkg.templates.{target_template.get_template_name_as_function_name()} or OID object or a list of the same"
        else:
            # component is not nested template, return normal documentation
            param_desc = f":param {self.get_clean_name()}: a parameter of type {self.value_class_label} or List[{self.value_class_label}] or OID object"
        if self.min_cardinality == 0:
            param_desc += " (optional)"
        return param_desc

    def create_object_statement(self) -> ObjectStatement:
        to_str = self.get_return_type() == "datetime.date"
        value = self.get_clean_name()
        statement = ObjectStatement(
            self.predicate_id, "UNKNOWN"
        )  # FIXME: template_id is unknown in this scope
        if not self.is_nested_template:
            if self.is_of_custom_type:
                statement.set_resource(
                    label=f"str({value})" if to_str else value,
                    classes=[self.value_class_id],
                )
            else:
                statement.set_literal(
                    text=f"str({value})" if to_str else value,
                    datatype=None,  # FIXME: datatype is not set for literals
                )
        return statement

    def __str__(self):
        return f'{"N" if self.is_nested_template else ""}Property(id="{self.predicate_id}", label="{self.predicate_label}", class="{self.value_class_id}", cardinality=({self.min_cardinality}, {self.max_cardinality}))'


class Template(object):
    # Static variables to keep track of all templates
    templates: Dict = {}
    wip_templates: Set = set()

    # Instance variables
    template_class: str
    is_formatted: bool
    template_id: str
    template_name: str
    components: List[TemplateComponent]
    is_strict: bool

    def __init__(self, orkg_client, template_id: str):
        # Add the template to the static variable
        Template.wip_templates.add(template_id)
        # Fetch template statements
        self.components = []
        template_statements = orkg_client.statements.get_by_subject(
            subject_id=template_id, size=99999
        ).content
        # Iterate over template statements and create template components
        components = filter(
            lambda x: x["predicate"]["id"] == "sh:property", template_statements
        )
        for component in components:
            self.components.append(
                TemplateComponent(orkg_client, component["object"]["id"])
            )
        # Set template class
        self.template_class = list(
            filter(
                lambda x: x["predicate"]["id"] == "sh:targetClass", template_statements
            )
        )[0]["object"]["id"]
        # Set template info
        self.template_id = template_id
        self.template_name = orkg_client.resources.by_id(id=template_id).content[
            "label"
        ]
        self.is_formatted = (
            len(
                list(
                    filter(
                        lambda x: x["predicate"]["id"] == "TemplateLabelFormat",
                        template_statements,
                    )
                )
            )
            > 0
        )
        strict_statement = list(
            filter(lambda x: x["predicate"]["id"] == "sh:closed", template_statements)
        )
        self.is_strict = (
            len(strict_statement) > 0
            and strict_statement[0]["object"]["label"].lower() == "true"
        )
        # Register template to the global templates dict
        Template.templates[template_id] = self
        # Sort components where the components with min cardinality 0 are at the end
        self.components.sort(key=lambda x: x.min_cardinality, reverse=True)

    @staticmethod
    def find_or_create_template(orkg_client, template_id: str) -> "Template":
        """
        Check if template is in registry, if yes return it.
        Otherwise, instantiate a new template.
        """
        if template_id in Template.templates:
            return Template.templates[template_id]
        else:
            return Template(orkg_client, template_id)

    def get_params_as_function_params(self) -> Tuple[str, str]:
        """
        Returns a tuple of the function parameters for the template.
        The first element is the function parameters as a string,
        the second element is the function docstring as a string.
        """
        params = ", ".join(
            [c.get_property_as_function_parameter() for c in self.components]
        )
        if not self.is_formatted:
            params = f"label: str, {params}"
            if params.strip()[-1] == ",":
                params = params.strip()[:-1]
        params_docstring = "\n\t".join(
            [comp.get_param_documentation() for comp in self.components]
        )
        if not self.is_formatted:
            params_docstring = f":param label: the label of the resource of type string\n\t{params_docstring}"
        return params, params_docstring

    def get_template_name_as_function_name(self) -> str:
        return check_against_keywords(pre_process_string(self.template_name))

    def create_api_object_values_and_classes(self) -> Tuple[List[ObjectStatement], str]:
        """
        Returns the API object values for the template.
        """
        properties = [
            c.create_object_statement()
            for c in self.components
            if c.value_class_id not in DATAFRAME_CLASSES
        ]
        return properties, f'"classes": ["{self.template_class}"]'

    def create_api_object(self) -> Tuple[str, List[ObjectStatement]]:
        """
        Returns a partial ORKG API object for the template as string., needs to be completed with the resource values.
        The second value returned is a list of the properties that are used by the code generator to fill the values
        """
        properties, classes = self.create_api_object_values_and_classes()
        object_json = f"""{{
                "resource": {{
                    "name": {'""' if self.is_formatted else 'label'},
                    {classes},
                    "values": {{}}
                }}
            }}"""
        return object_json, properties

    def __str__(self):
        return f'Template(id="{self.template_id}", class="{self.template_class}", components= "{len(self.components)} comps."))'



from orkg.out import OrkgResponse
from orkg.utils import NamespacedClient


class StatsClient(NamespacedClient):
    def get(self) -> OrkgResponse:
        self.client.backend._append_slash = True
        response = self.client.backend.stats.GET()
        return self.client.wrap_response(response)



from typing import Dict, List, Optional, Union

from orkg.common import ThingType
from orkg.out import OrkgResponse
from orkg.utils import NamespacedClient, simcomp_available


class JSONClient(NamespacedClient):
    @simcomp_available
    def save_json(
        self,
        thing_key: str,
        thing_type: ThingType = ThingType.UNKNOWN,
        data: Optional[Union[Dict, List]] = None,
        config: Optional[Dict] = None,
    ) -> OrkgResponse:
        """
        Save a json object to the simcomp relational database
        :param thing_key: the key of the thing
        :param thing_type: the type of the thing (UNKNOWN, COMPARISON, DIAGRAM, VISUALIZATION, DRAFT_COMPARISON, LIST, REVIEW, QUALITY_REVIEW, PAPER_VERSION, ANY)
        :param data: the json object to save
        :param config: the config of the thing, used for comparisons mostly
        :return: the response of the request
        """
        # Check if the thing already exists
        res = self.client.json.get_json(thing_key, thing_type)
        if res.succeeded:
            raise ValueError(
                f"thing with key {thing_key} and type {thing_type.value} already exists!"
            )
        # Save the thing
        if config is None:
            config = {}
        if data is None:
            data = {}
        return self.client.wrap_response(
            self.client.simcomp.thing.POST(
                json={
                    "thing_key": thing_key,
                    "data": data,
                    "thing_type": thing_type.value,
                    "config": config,
                }
            )
        )

    @simcomp_available
    def get_json(self, thing_key: str, thing_type: ThingType) -> OrkgResponse:
        """
        Get a json object from the simcomp relational database
        :param thing_key: the key of the thing
        :param thing_type: the type of the thing
        :return: the response of the request
        """
        return self.client.wrap_response(
            self.client.simcomp.thing.GET(
                params={"thing_key": thing_key, "thing_type": thing_type.value}
            )
        )
